/*********************************************************************
 *	ATtiny2313 Serial Port Bootloader
 *********************************************************************
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "loader.h"
//#include "hidasp.h"
#include "util.h"



#define DEFAULT_SERIAL	"0001"

char HELP_USAGE[]={
	"* tinyLoader Ver 0.1 (" __DATE__ ")\n"
	"Usage is\n"
	"  tinyLoader.exe  [Options] hexfile.hex\n"
	"Options\n"
	"  -rp         ...  Read Flash.\n"
	"  -w          ...  Write Flash.\n"
	"  -v          ...  Verify Flash.\n"
	"  -r          ...  Boot.\n"
"\n"
	"  -p<comPortNum>  ...  Select serial port number.\n"
	"  -b<BaudRate>    ...  Select baudrate(Default=19200).\n"
};

#define	MAX_CMDBUF	4096

static	char lnbuf[1024];
static	char verbose_mode = 0;	//1:foCX_v.
		int  hidmon_mode = 1;	/* hidmon s */
static	uchar 	databuf[256];
static	int		dataidx;
static	int		adr_u = 0;		// hexrec mode 4


#define	FLASH_START		0x0
#define	FLASH_SIZE		0x800
#define	FLASH_SIZE_USER	0x6c0
#define	FLASH_STEP		32

#define	FLASH_OFFSET 0		// 0x800 = eXg.

static	uchar flash_buf[FLASH_SIZE];
static	int	  flash_start = FLASH_START;
static	int	  flash_end   = FLASH_SIZE_USER;

static	char  opt_r  = 0;	//	'-r '
static	char  opt_rp = 0;	//	'-rp'
static	char  opt_re = 0;	//	'-re'
static	char  opt_rf = 0;	//	'-rf'

#define	CHECKSUM_CALC	(1)


static  char *com_port = "9";	// "COM9:" ftHg.
static  int   baudrate = 19200;

/*********************************************************************
 *	gp@\
 *********************************************************************
 */
void usage(void)
{
	fputs( HELP_USAGE, stdout );
}
/*********************************************************************
 *	s̓[eBeB
 *********************************************************************
 */
static	int getln(char *buf,FILE *fp)
{
	int c;
	int l;
	l=0;
	while(1) {
		c=getc(fp);
		if(c==EOF)  {
			*buf = 0;
			return(EOF);/* EOF */
		}
		if(c==0x0d) continue;
		if(c==0x0a) {
			*buf = 0;	/* EOL */
			return(0);	/* OK  */
		}
		*buf++ = c;l++;
		if(l>=255) {
			*buf = 0;
			return(1);	/* Too long line */
		}
	}
}
/**********************************************************************
 *	INTEL HEXR[h `01 o͂.
 **********************************************************************
 */
static	int	out_ihex01(unsigned char *data,int adrs,int cnt,FILE *ofp)
{
	unsigned char buf[1024];
	buf[0] = cnt;
	buf[1] = adrs >> 8;
	buf[2] = adrs & 0xff;
	buf[3] = 0;
	int length = cnt+5;
	int i;
	int sum=0;

	memcpy(buf+4,data,cnt);
	for(i=0;i<length-1;i++) {
		sum += buf[i];
	}
	sum = 256 - (sum&0xff);
	buf[length-1] = sum;

	fprintf(ofp,":");
	for(i=0;i<length;i++) {
		fprintf(ofp,"%02X",buf[i]);
	}
	fprintf(ofp,"\n");
	return 0;
}
/**********************************************************************
 *	INTEL HEXR[h `01 (oCi[ς) ߂.
 **********************************************************************
 */
static	int	parse_ihex01(unsigned char *buf,int length)
{
	
	int adrs = (buf[1] << 8) | buf[2];
	int cnt  =  buf[0];
	int i;
	int sum=0;

#if	CHECKSUM_CALC
	for(i=0;i<length;i++) {
		sum += buf[i];
	}
//	printf("checksum=%x\n",sum);
	if( (sum & 0xff) != 0 ) {
		printf("HEX RECORD Checksum Error!\n");
		exit(EXIT_FAILURE);
	}
#endif

	buf += 4;

#if	HEX_DUMP_TEST
	printf("%04x",adrs|adr_u);
	for(i=0;i<cnt;i++) {
		printf(" %02x",buf[i]);
	}
	printf("\n");
#endif


	if(adr_u != 0) return 1;	//ʃAhX[(64kBȊO)

	for(i=0;i<cnt;i++) {
		flash_buf[adrs+i] = buf[i];
	}
	return 0;
}
/**********************************************************************
 *	INTEL HEXR[h(Ps:oCi[ς) ߂.
 **********************************************************************
 */
int	parse_ihex(int scmd,unsigned char *buf,int length)
{
	switch(scmd) {
		case 0x00:	//f[^R[h:
					//f[^tB[h͏܂ׂf[^łB
			return parse_ihex01(buf,length);

		case 0x01:	//GhR[h:
					//HEXt@C̏I.t񖳂.
			return scmd;
			break;

		case 0x02:	//ZOgAhXR[h:
					//f[^tB[hɂ̓ZOgAhXB Ƃ΁A
					//:02000002E0100C
					//		   ~~~~
					//
			break;

		case 0x03:	//X^[gZOgAhXR[h.
			break;

		case 0x04:	//gjAAhXR[h.
					//̃R[hł32rbgAhX̂16rbg
					//irbg32`rbg16́j^B
				adr_u = (buf[4]<<24)|(buf[5]<<16);
//				printf("adr_u=%x\n",adr_u);
				return scmd;
			break;

		case 0x05:	//X^[gjAAhXF
					//Ⴆ
					//:04000005FF000123D4
					//         ~~~~~~~~
					//ƂȂĂ΁AFF000123hԒnX^[gAhXɂȂB

			break;
		
	}
	printf("Unsupported ihex cmd=%x\n",scmd);
	return 0;
}

/**********************************************************************
 *	16i2̕oCi[ɕϊ.
 **********************************************************************
 */
int read_hex_byte(char *s)
{
	char buf[16];
	int rc = -1;

	buf[0] = s[0];
	buf[1] = s[1];
	buf[2] = 0;
	sscanf(buf,"%x",&rc);
	return rc;
}

/**********************************************************************
 *	16iŏꂽoCi[ɕϊ.
 **********************************************************************
 */
static	int read_hex_string(char *s)
{
	int rc;

	dataidx = 0;
	while(*s) {
		rc = read_hex_byte(s);s+=2;
		if(rc<0) return rc;
		databuf[dataidx++]=rc;
	}
	return 0;
}

/**********************************************************************
 *		
 **********************************************************************
 0 1 2 3 4
:1000100084C083C082C081C080C07FC07EC07DC0DC
 ~~ f[^
   ~~~~ AhX
       ~~ R[h^Cv
         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~f[^ 
                                         ~~`FbNT.
 */
int	read_ihex(char *s)
{
	int c;
	if(s[0] == ':') {
		read_hex_string(s+1);
		c = databuf[3];
		parse_ihex(c,databuf,dataidx);
	}
	return 0;
}
/*********************************************************************
 *	HEX file ǂݍ flash_buf[] Ɋi[.
 *********************************************************************
 */
int	read_hexfile(char *fname)
{
	FILE *ifp;
	int rc;
	Ropen(fname);
	while(1) {
		rc = getln(lnbuf,ifp);
		read_ihex(lnbuf);
		if(rc == EOF) break;
	}
	Rclose();
	return 0;
}

/*********************************************************************
 *	flash_buf[] Ɋi[	ZbgxN^̒u.
 *********************************************************************
 */
int	modify_flash_buf(void)
{
	if( (flash_buf[0]== 0x5f) && (flash_buf[1]== 0xc3) ) {
		//łɃZbgxN^[̓u[g[_[ɐUĂ.
	}else{
		//ZbgxN^[ rjmp ߂Ɖ肷.
		//ȊOɂ͍̂ƂΉłȂ.
		int inst0 = (flash_buf[1]<<8) | flash_buf[0];
		if( (inst0 & 0xc000)!=0xc000 ) {
			printf("Fatal: reset vector error!(not rjmp) [%x]\n",inst0);
		}
		int jmpvec = 0x0002 + ((inst0 & 0x0fff)<<1);	// rjmp̊_͎̖߁0002 Ԓn.


		// ZbgxN^[̏.
		//00000000 <__vectors>:
		//   0:	5f c3       	rjmp	.+1726   	; 0x6c0 <__init>
		flash_buf[0]= 0x5f;
		flash_buf[1]= 0xc3;

		int inst1 = ( (jmpvec-0x6c0) >>1 ) & 0xfff;	// offset12bit
		inst1 |= 0xc000;		// rjmp
		// [U[ZbgxN^[̑.
		//000006be <user_proc>:
		//   06be:	5f c3     	rjmp	
		flash_buf[0x6be]= inst1 & 0xff;
		flash_buf[0x6bf]= inst1 >> 8;
	}
	return 0;
}
/*********************************************************************
 *	^[Qbg̍ċN.
 *********************************************************************
 */
int reboot_target(void)
{
	UsbBootTarget(0,1);
	return 0;
}

/*********************************************************************
 *	flash_buf[] ̓e ^[Qbg AVR ɏ.
 *********************************************************************
 */
int	write_block_for_test(int addr,int size)
{
	int i,f=0;
	for(i=0;i<size;i++) {
		if(flash_buf[addr+i]!=0xff) f=1;
	}

	if(f) {
		printf("%04x",addr);
		for(i=0;i<size;i++) {
			printf(" %02x",flash_buf[addr+i]);
		}
		printf("\n");
	}
	return 0;
}
/*********************************************************************
 *	64byte.
 *********************************************************************
 */
int	erase_block(int addr,int size)
{
	int i,f=0;
	int pages = size / 64;
	
	for(i=0;i<size;i++) {
		if(flash_buf[addr+i]!=0xff) f=1;
	}

	if(f) {
		UsbEraseTargetROM(addr + FLASH_OFFSET,pages);
		Sleep(25);
	}
	return 0;
}
/*********************************************************************
 *	flash_buf[] ̓e ^[Qbg AVR ɏ.
 *********************************************************************
 */
int	write_block(int addr,int size)
{
	int i,f=0;
	for(i=0;i<size;i++) {
		if(flash_buf[addr+i]!=0xff) f=1;
	}

	if(f) {
		UsbFlash(addr + FLASH_OFFSET,AREA_PGMEM,flash_buf+addr,size);
		Sleep(3);
		printf("Writing ... %04x\r",addr);
		for(i=0;i<size;i++) {
//			printf(" %02x",flash_buf[addr+i]);
		}
//		printf("\n");
		

	}
	return 0;
}
/*********************************************************************
 *	flash_buf[] ̓e ^[Qbg AVR ɏ.
 *********************************************************************
 */
int	verify_block(int addr,int size)
{
	int i,f=0;
	int errcnt=0;
	uchar 	verify_buf[256];
	for(i=0;i<size;i++) {
		if(flash_buf[addr+i]!=0xff) f=1;
	}

	if(f) {
		UsbRead(addr + FLASH_OFFSET,AREA_PGMEM,verify_buf,size);
		for(i=0;i<size;i++) {
			if(flash_buf[addr+i] != verify_buf[i]) {
				errcnt++;
				printf("Verify Error in %04x : write %02x != read %02x\n"
					,addr+i,flash_buf[addr+i],verify_buf[i]);
			}
		}
		printf("Verifying ... %04x\r",addr);
	}
	return errcnt;
}
/*********************************************************************
 *	 ^[Qbg AVR ̓e flash_buf[] ɓǂݍ.
 *********************************************************************
 */
int	read_block(int addr,int size)
{
	int errcnt=0;
	uchar *read_buf;
	
	read_buf = &flash_buf[addr] ;
	UsbRead(addr + FLASH_OFFSET,AREA_PGMEM,read_buf,size);
//	memdump("PGMEM",read_buf,size);
	return errcnt;
}
/*********************************************************************
 *	_~[.
 *********************************************************************
 */
int disasm_print(unsigned char *buf,int size,int adr)
{
	unsigned short *inst = (unsigned short *)buf;
	printf("%04x %04x\n",adr,inst[0]);
	return 2;
}
/*********************************************************************
 *	flash_buf[] ̓e ^[Qbg AVR ɏ.
 *********************************************************************
 */
int	write_hexdata(void)
{
	int i;
	for(i=flash_start;i<flash_end;i+= FLASH_STEP) {
		write_block(i,FLASH_STEP);
	}
	return 0;
}
/*********************************************************************
 *	flash_buf[] ̓e ^[Qbg AVR ɏ.
 *********************************************************************
 */
int	verify_hexdata(void)
{
	int i;
	int errcnt = 0;
	printf("\n");
	for(i=flash_start;i<flash_end;i+= FLASH_STEP) {
		errcnt += verify_block(i,FLASH_STEP);
	}
	return errcnt;
}
/*********************************************************************
 *	|[ĝAhX߂.i_~[j
 *********************************************************************
 */
int	portAddress(char *s)
{
	return 0;
}

/*********************************************************************
 *	ROMǂݏoʂHEXŏo͂.
 *********************************************************************
 */
#define	HEX_DUMP_STEP	16

void print_hex_block(FILE *ofp,int addr,int size)
{
	int i,j,f;
	for(i=0;i<size;i+=HEX_DUMP_STEP,addr+=HEX_DUMP_STEP) {
		f = 0;
		for(j=0;j<HEX_DUMP_STEP;j++) {
			if( flash_buf[addr+j] != 0xff ) {
				f = 1;
			}
		}
		if(f) {

#if	0
			printf(":%04x",addr);
			for(j=0;j<HEX_DUMP_STEP;j++) {
				printf(" %02x",flash_buf[addr+j]);
			}
			printf("\n");
#endif
			out_ihex01(&flash_buf[addr],addr,HEX_DUMP_STEP,ofp);


		}
	}
	

}
/*********************************************************************
 *	ROMǂݏo.
 *********************************************************************
 */
void read_from_avr(char *file)
{
	int i;
	FILE *ofp;
	fprintf(stderr, "Reading...\n");
	Wopen(file);
	fprintf(ofp,":020000040000FA\n");
	for(i=flash_start;i<flash_end;i+= FLASH_STEP) {
		printf("%04x\r",i);
		read_block(i,FLASH_STEP);
		print_hex_block(ofp,i,FLASH_STEP);
		//printf("%x\n",i);
	}
	fprintf(ofp,":00000001FF\n");
	Wclose();
}
/*********************************************************************
 *	C
 *********************************************************************
 */
int main(int argc,char **argv)
{
	int errcnt, ret_val;

	//IvV.
	Getopt(argc,argv,"i");
	if(IsOpt('h') || IsOpt('?') || IsOpt('/')) {
		usage();
		exit(EXIT_SUCCESS);
	}
	if(IsOpt('c')) {
		com_port = Opt('c');		// COM Port No.
	}
	if(IsOpt('b')) {
		sscanf(Opt('b'),"%d",&baudrate); // Get BaudRate
//		printf("%d\n",baudrate);
	}
	if((argc<2)&& (!IsOpt('r')) && (!IsOpt('p')) ) {
		usage();
		exit(EXIT_SUCCESS);
	}
	if(IsOpt('B')) {
		flash_start = 0;
		flash_end   = FLASH_SIZE;
	}

	if(IsOpt('r')) {
		char *r = Opt('r');
		if(r[0]==0  ) opt_r = 1;
		if(r[0]=='p') opt_rp = 1;
		if(r[0]=='e') opt_re = 1;
		if(r[0]=='f') opt_rf = 1;
	}

	//.
	if( UsbInit(verbose_mode, baudrate, com_port ) ) {
		exit(EXIT_FAILURE);
	}

	memset(flash_buf,0xff,FLASH_SIZE);
	// ZbgxN^[̏.
	//00000000 <__vectors>:
	//   0:	5f c3       	rjmp	.+1726   	; 0x6c0 <__init>
	flash_buf[0]= 0x5f;
	flash_buf[1]= 0xc3;


	ret_val = EXIT_SUCCESS;

	if(argc>=2) {
		if(opt_rp||opt_re||opt_rf) {	// ROMǂݏo.
			read_from_avr(argv[1]);
		}else{
			read_hexfile(argv[1]);		//	HEXt@C̓ǂݍ.
			modify_flash_buf();			// ZbgxN^̒u.

			if(IsOpt('v')==0) {			// xt@ĈƂ͏݂Ȃ.
				write_hexdata();			//  .
			}

			errcnt = verify_hexdata();	//  xt@C.
			if(errcnt==0) {
				printf("Verify Ok.\n");
			}else{
				printf("Verify Error. errors=%d\n",errcnt);
				ret_val = EXIT_FAILURE;
			}
		}
	}

	if(opt_r) {
		reboot_target();			//  ^[QbgċN.
	}
	UsbExit();
	return ret_val;
}
/*********************************************************************
 *
 *********************************************************************
 */

